@@ -4,15 +4,15 @@ use bytes::Bytes;
 use clap::Clap;
 use console::style;
 use exocore_chain::{
-    block::{Block, BlockBuilder, BlockOperations},
-    chain::ChainStore,
+    block::{Block, BlockBuilder, BlockOperations, DataBlock},
+    chain::{ChainData, ChainStore},
     operation::{OperationBuilder, OperationFrame, OperationId},
     DirectoryChainStore, DirectoryChainStoreConfig,
 };
 use exocore_core::{
     cell::{
-        Cell, CellConfigExt, CellId, CellNode, CellNodeRole, EitherCell, FullCell, LocalNode,
-        LocalNodeConfigExt, NodeConfigExt,
+        Cell, CellConfigExt, CellId, CellNode, CellNodeConfigExt, CellNodeRole, EitherCell,
+        FullCell, LocalNode, LocalNodeConfigExt,
     },
     framing::{sized::SizedFrameReaderIterator, FrameReader},
     sec::{auth_token::AuthToken, keys::Keypair},
@@ -21,7 +21,7 @@ use exocore_core::{
 use exocore_protos::{
     core::{
         cell_application_config, cell_node_config, node_cell_config, CellConfig, CellNodeConfig,
-        LocalNodeConfig, NodeCellConfig, NodeConfig,
+        LocalNodeConfig, NodeCellConfig,
     },
     generated::data_chain_capnp::{block_header, chain_operation},
     prost::Message,
@@ -31,9 +31,8 @@ use exocore_protos::{
 use crate::{
     app::AppPackage,
     disco::prompt_discovery_pin,
-    term::*,
     utils::{edit_file, edit_string},
-    Context,
+    Context, *,
 };
 
 #[derive(Clap, Clone)]
@@ -54,10 +53,10 @@ pub struct CellOptions {
 
 #[derive(Clap, Clone)]
 enum CellCommand {
-    /// Initialize a new cell.
+    /// Initializes a new cell.
     Init(InitOptions),
 
-    /// List cells of the node.
+    /// Lists cells of the node.
     List,
 
     /// Edit cell configuration.
@@ -69,26 +68,38 @@ enum CellCommand {
     /// Cell nodes related commands.
     Node(NodeOptions),
 
-    /// Print cell configuration.
+    /// Prints cell configuration.
     Print(PrintOptions),
 
     /// Cell apps related commands.
     App(AppOptions),
 
-    /// Check the cell's chain integrity.
-    CheckChain,
+    /// Cell chain related commands.
+    Chain(CellChainOptions),
 
-    /// Export the chain's data.
-    ExportChain(ChainExportOptions),
-
-    /// Import the chain's data.
-    ImportChain(ChainImportOptions),
-
-    /// Generate an auth token.
+    /// Generates an auth token.
     GenerateAuthToken(GenerateAuthTokenOptions),
+}
 
-    /// Create genesis block of the chain.
-    CreateGenesisBlock,
+#[derive(Clap, Clone)]
+struct CellChainOptions {
+    #[clap(subcommand)]
+    command: CellChainCommand,
+}
+
+#[derive(Clap, Clone)]
+enum CellChainCommand {
+    /// Initializes a chain with a genesis block.
+    Init,
+
+    /// Checks the cell's chain integrity.
+    Check,
+
+    /// Exports the chain's data.
+    Export(ChainExportOptions),
+
+    /// Imports the chain's data.
+    Import(ChainImportOptions),
 }
 
 #[derive(Clap, Clone)]
@@ -110,7 +121,7 @@ struct InitOptions {
     #[clap(long)]
     no_app_host: bool,
 
-    /// Don't create genesis block.
+    /// Don't initialize the chain with a genesis block.
     #[clap(long)]
     no_genesis: bool,
 }
@@ -118,6 +129,18 @@ struct InitOptions {
 /// Cell join related options
 #[derive(Clap, Clone)]
 struct JoinOptions {
+    /// The node will host the chain locally.
+    #[clap(long)]
+    chain: bool,
+
+    /// The node will host entities store.
+    #[clap(long)]
+    store: bool,
+
+    /// The node will host applications.
+    #[clap(long)]
+    app_host: bool,
+
     /// Manually join a cell using its cell configuration yaml.
     #[clap(long)]
     manual: bool,
@@ -151,18 +174,6 @@ enum NodeCommand {
 
 #[derive(Clap, Clone)]
 struct NodeAddOptions {
-    /// The node will host the chain locally.
-    #[clap(long)]
-    chain: bool,
-
-    /// The node will host entities store.
-    #[clap(long)]
-    store: bool,
-
-    /// The node will host applications.
-    #[clap(long)]
-    app_host: bool,
-
     /// Manually add the node using its node configuration yaml.
     #[clap(long)]
     manual: bool,
@@ -271,14 +282,16 @@ pub async fn handle_cmd(ctx: &Context, cell_opts: &CellOptions) -> anyhow::Resul
             cmd_print(ctx, cell_opts, opts);
             Ok(())
         }
-        CellCommand::CheckChain => cmd_check_chain(ctx, cell_opts),
-        CellCommand::ExportChain(export_opts) => cmd_export_chain(ctx, cell_opts, export_opts),
-        CellCommand::ImportChain(import_opts) => cmd_import_chain(ctx, cell_opts, import_opts),
+        CellCommand::Chain(chain_opts) => match &chain_opts.command {
+            CellChainCommand::Init => cmd_create_genesis_block(ctx, cell_opts),
+            CellChainCommand::Check => cmd_check_chain(ctx, cell_opts),
+            CellChainCommand::Export(export_opts) => cmd_export_chain(ctx, cell_opts, export_opts),
+            CellChainCommand::Import(import_opts) => cmd_import_chain(ctx, cell_opts, import_opts),
+        },
         CellCommand::GenerateAuthToken(gen_opts) => {
             cmd_generate_auth_token(ctx, cell_opts, gen_opts);
             Ok(())
         }
-        CellCommand::CreateGenesisBlock => cmd_create_genesis_block(ctx, cell_opts),
     }
 }
 
@@ -308,26 +321,15 @@ fn cmd_init(
             .interact_text()?;
     }
 
+    // Create a configuration for the node in the cell
     let cell_node = {
-        // Create a configuration for the node in the cell
-        let mut cell_node = CellNodeConfig {
-            node: Some(NodeConfig {
-                public_key: node.public_key().encode_base58_string(),
-                id: node.id().to_string(),
-                name: node.name().to_string(),
-                addresses: node_config.addresses.clone(),
-            }),
-            ..Default::default()
-        };
-
+        let mut roles = Vec::new();
         if !init_opts.no_chain {
             print_action(format!(
                 "The node will have {} role",
                 style_emphasis("chain")
             ));
-            cell_node
-                .roles
-                .push(cell_node_config::Role::ChainRole.into());
+            roles.push(cell_node_config::Role::ChainRole);
         }
 
         if !init_opts.no_store {
@@ -335,9 +337,7 @@ fn cmd_init(
                 "The node will have {} role",
                 style_emphasis("store")
             ));
-            cell_node
-                .roles
-                .push(cell_node_config::Role::StoreRole.into());
+            roles.push(cell_node_config::Role::StoreRole);
         }
 
         if !init_opts.no_app_host {
@@ -345,12 +345,10 @@ fn cmd_init(
                 "The node will have {} role",
                 style_emphasis("app host")
             ));
-            cell_node
-                .roles
-                .push(cell_node_config::Role::AppHostRole.into());
+            roles.push(cell_node_config::Role::AppHostRole);
         }
 
-        cell_node
+        node_config.create_cell_node_config(roles)
     };
 
     let cell_config = {
@@ -409,15 +407,15 @@ async fn cmd_node_add(
 
     let disco_client = ctx.get_discovery_client();
 
-    let (new_node_config, reply_pin, reply_token) = if add_opts.manual {
-        let node_config = edit_string(
-            "# Paste joining node's public info (result of `exo config print --cell` on joining node)",
+    let (cell_node, reply_pin, reply_token) = if add_opts.manual {
+        let cell_node_ = edit_string(
+            "# Paste joining node's public info (result of `exo cell join --manual [--role, [--role, ...]]` on joining node)",
             |config| {
-                let config = NodeConfig::from_yaml(config.as_bytes())?;
+                let config = CellNodeConfig::from_yaml(config.as_bytes())?;
                 Ok(config)
             },
         );
-        (node_config, None, None)
+        (cell_node_, None, None)
     } else {
         print_spacer();
         let pin = prompt_discovery_pin(ctx, "Enter the joining node discovery PIN");
@@ -426,48 +424,58 @@ async fn cmd_node_add(
             .await
             .expect("Couldn't find joining node using the given discovery pin");
 
-        let node_config_yml = payload
+        let cell_node_yml = payload
             .decode_payload()
             .expect("Couldn't decode node payload");
-        let node_config = NodeConfig::from_yaml(node_config_yml.as_slice())
+        let cell_node = CellNodeConfig::from_yaml(cell_node_yml.as_slice())
             .expect("Couldn't parse joining node config");
 
-        (node_config, payload.reply_pin, payload.reply_token)
+        (cell_node, payload.reply_pin, payload.reply_token)
     };
 
-    let mut cell_node = CellNodeConfig {
-        node: Some(new_node_config.clone()),
-        roles: vec![],
+    let node_config = cell_node.node.clone().unwrap();
+
+    let has_role = |role: cell_node_config::Role| -> bool {
+        let role_i32: i32 = role.into();
+        for another_role in &cell_node.roles {
+            if role_i32 == *another_role {
+                return true;
+            }
+        }
+        false
     };
 
-    if add_opts.chain {
-        print_action(format!(
-            "The node will have {} role",
-            style_emphasis("chain")
-        ));
-        cell_node
-            .roles
-            .push(cell_node_config::Role::ChainRole.into());
+    print_step("New node found !");
+
+    print_info(format!("Node name: {}", style_value(&node_config.name)));
+    print_info(format!(
+        "Public key: {}",
+        style_value(&node_config.public_key)
+    ));
+    print_info(format!(
+        "Addresses: {}",
+        style_value(&node_config.addresses.clone().unwrap_or_default())
+    ));
+
+    if !cell_node.roles.is_empty() {
+        print_info("Roles:");
+        if has_role(cell_node_config::Role::ChainRole) {
+            print_action(style_emphasis("chain"));
+        }
+
+        if has_role(cell_node_config::Role::StoreRole) {
+            print_action(style_emphasis("store"));
+        }
+
+        if has_role(cell_node_config::Role::AppHostRole) {
+            print_action(style_emphasis("application host"));
+        }
+    } else {
+        print_info("The node will have no roles");
     }
 
-    if add_opts.store {
-        print_action(format!(
-            "The node will have {} role",
-            style_emphasis("store")
-        ));
-        cell_node
-            .roles
-            .push(cell_node_config::Role::StoreRole.into());
-    }
-
-    if add_opts.app_host {
-        print_action(format!(
-            "The node will have {} role",
-            style_emphasis("application host")
-        ));
-        cell_node
-            .roles
-            .push(cell_node_config::Role::AppHostRole.into());
+    if !confirm(ctx, "Do you want to add the node to the cell?") {
+        return Err(anyhow!("Operation aborted"));
     }
 
     cell_config.add_node(cell_node);
@@ -480,13 +488,13 @@ async fn cmd_node_add(
         .to_yaml_file(&config_path)
         .expect("Couldn't write cell config");
 
-    if !add_opts.manual {
-        let cell_config_inlined = cell_config
-            .inlined()
-            .expect("Couldn't inline cell config")
-            .to_yaml()
-            .expect("Couldn't convert cell config to yaml");
+    let cell_config_inlined = cell_config
+        .inlined()
+        .expect("Couldn't inline cell config")
+        .to_yaml()
+        .expect("Couldn't convert cell config to yaml");
 
+    if !add_opts.manual {
         disco_client
             .reply(
                 reply_pin.expect("Expected reply pin, but didn't find one"),
@@ -499,8 +507,10 @@ async fn cmd_node_add(
 
         print_success(format!(
             "Node {} has been added to the cell",
-            style_value(new_node_config.name)
+            style_value(node_config.name)
         ));
+    } else if confirm(ctx, "Do you want to print cell configuration?") {
+        print!("{}", cell_config_inlined);
     }
 
     Ok(())
@@ -551,12 +561,33 @@ async fn cmd_join(
     let node_config = ctx.options.read_configuration();
     let disco_client = ctx.get_discovery_client();
 
-    let cell_node = NodeConfig {
-        id: node_config.id.clone(),
-        name: node_config.name.clone(),
-        public_key: node_config.public_key.clone(),
-        addresses: node_config.addresses.clone(),
-    };
+    let mut roles = Vec::new();
+
+    if join_opts.chain {
+        print_action(format!(
+            "The node will have {} role",
+            style_emphasis("chain")
+        ));
+        roles.push(cell_node_config::Role::ChainRole);
+    }
+
+    if join_opts.store {
+        print_action(format!(
+            "The node will have {} role",
+            style_emphasis("store")
+        ));
+        roles.push(cell_node_config::Role::StoreRole);
+    }
+
+    if join_opts.app_host {
+        print_action(format!(
+            "The node will have {} role",
+            style_emphasis("app host")
+        ));
+        roles.push(cell_node_config::Role::AppHostRole);
+    }
+
+    let cell_node = node_config.create_cell_node_config(roles);
     let cell_node_yaml = cell_node
         .to_yaml()
         .expect("Couldn't convert cell node config to yaml");
@@ -571,6 +602,7 @@ async fn cmd_join(
             "On the host node, enter this discovery pin:\n\n\t\t{}",
             style_value(create_resp.pin.to_formatted_string())
         ));
+        print_spacer();
 
         let payload = disco_client
             .get_loop(create_resp.reply_pin.unwrap(), Duration::from_secs(60))
@@ -583,6 +615,11 @@ async fn cmd_join(
         CellConfig::from_yaml(cell_config_yml.as_slice())
             .expect("Couldn't parse cell config from host node")
     } else {
+        print_info("Paste node cell information on host node:");
+        println!("{}", cell_node_yaml);
+
+        wait_press_enter();
+
         edit_string(
             "# Paste config of the cell to join (result of `exo cell print --inline` on host node)",
             |config| {
@@ -686,17 +723,19 @@ fn cmd_check_chain(ctx: &Context, cell_opts: &CellOptions) -> anyhow::Result<()>
     print_spacer();
     let bar = indicatif::ProgressBar::new(last_block.get_height()?);
 
+    let mut prev_block: Option<DataBlock<ChainData>> = None;
     let mut block_count = 0;
     for block in chain_store.blocks_iter(0) {
         let block = block?;
 
         block_count += 1;
-        if let Err(err) = block.validate() {
+        if let Err(err) = block.validate(prev_block) {
             let block_header_reader = block.header().get_reader();
             let block_height = block_header_reader
                 .map(block_header::Reader::get_height)
                 .ok();
 
+            bar.finish_and_clear();
             print_error(format!(
                 "Block at offset={} height={} is invalid: {}",
                 style_value(block.offset()),
@@ -706,6 +745,8 @@ fn cmd_check_chain(ctx: &Context, cell_opts: &CellOptions) -> anyhow::Result<()>
             return Ok(());
         }
 
+        prev_block = Some(block);
+
         bar.set_position(block_count);
     }
 
@@ -850,13 +891,7 @@ fn cmd_import_chain(
             style_value(last_block.get_height()),
         ));
 
-        print_spacer();
-        let over = dialoguer::Confirm::with_theme(ctx.dialog_theme.as_ref())
-            .with_prompt("Do you want to wipe the chain?")
-            .interact()
-            .expect("Couldn't get prompt answer");
-
-        if over {
+        if confirm(ctx, "Do you want to wipe the chain?") {
             chain_store.truncate_from_offset(0)?;
         } else {
             panic!("Chain is already initialized");
@@ -976,10 +1011,9 @@ fn cmd_generate_auth_token(
         "Expiration: {}",
         style_value(expiration.to_datetime())
     ));
-    print_info(format!(
-        "Token: {}",
-        style_value(token.encode_base58_string())
-    ));
+
+    print_info("Token:");
+    println!("{}", token.encode_base58_string());
 }
 
 fn cmd_create_genesis_block(ctx: &Context, cell_opts: &CellOptions) -> anyhow::Result<()> {
@@ -1163,7 +1197,7 @@ fn extract_cell_by_name(either_cells: Vec<EitherCell>, name: &str) -> Option<Eit
     either_cells.into_iter().find(|c| c.cell().name() == name)
 }
 
-fn cell_config_path(cell: &Cell) -> PathBuf {
+pub fn cell_config_path(cell: &Cell) -> PathBuf {
     let cell_directory = cell.cell_directory().expect("Couldn't find cell directory");
     cell_directory.join("cell.yaml")
 }
@@ -1199,7 +1233,7 @@ fn create_genesis_block(cell: FullCell) -> anyhow::Result<()> {
     Ok(())
 }
 
-fn write_cell_config(ctx: &Context, config: &CellConfig) {
+pub fn write_cell_config(ctx: &Context, config: &CellConfig) {
     if config.public_key.is_empty() {
         panic!("Expected cell to have a public key");
     }
@@ -1244,3 +1278,29 @@ fn add_node_config_cell(ctx: &Context, node_config: &LocalNodeConfig, cell_confi
         .to_yaml_file(ctx.options.conf_path())
         .expect("Couldn't write node config");
 }
+
+pub fn copy_local_node_to_cells(ctx: &Context, node_config: LocalNodeConfig) {
+    let (either_cells, _local_node) = Cell::from_local_node_config(node_config.clone())
+        .expect("Couldn't create cell from config");
+
+    for cell in either_cells {
+        let config_path = cell_config_path(cell.cell());
+        let mut cell_config =
+            CellConfig::from_yaml_file(config_path).expect("Couldn't read cell config");
+
+        let changed = if let Some(cell_node_config) = cell_config
+            .find_node(&node_config.public_key)
+            .and_then(|c| c.node.as_mut())
+        {
+            cell_node_config.name = node_config.name.clone();
+            cell_node_config.addresses = node_config.addresses.clone();
+            true
+        } else {
+            false
+        };
+
+        if changed {
+            write_cell_config(ctx, &cell_config);
+        }
+    }
+}
